unit	UCmdLine;


//	Parameters whose first character is not a "/" are arguments.
//	Parameters whose first character is     a "/" are switches.
//	Switches can have options, for example "/switch=option1,option2,option3".
//
//	"m_Arguments" and "m_Switches" are implemented as stringlists. "m_Options" is implemented
//	as a list of stringlists, i.e. "m_Options[4]" contains the options for switch #4 in its
//	own stringlist.
//
//	Switches are internally stored in uppercase.


interface  ////////////////////////////////////////////////////////////////////////////////////
uses	Windows, Classes,
	UStringLists;


type	TCmdLine	=	class
				private
				m_Arguments		:	TStringList;
				m_Options		:	TStringLists;  // list of lists
				m_Switches		:	TStringList;
				private
				function    Get_Argument	(const Index  : Word) : String;
				function    Get_Option		(const Switch : String;  const Index : Word) : String;
				function    Get_OptionCount	(const Switch : String) : Integer;
				procedure   Parse;
				public
				constructor Create		;
				destructor  Destroy		;  override;
				function    has_Switch		(const Switch : String) : Boolean;
				public
				property    Arguments		[const Index  : Word  ] : String   read Get_Argument;
				property    OptionCount		[const Switch : String] : Integer  read Get_OptionCount;
				property    Options		[const Switch : String;  const Index : Word] : String  read Get_Option;
				end;


var	CmdLine		:	TCmdLine;


implementation  ///////////////////////////////////////////////////////////////////////////////
uses	SysUtils;


const	CRLF		=	#13#10;


constructor TCmdLine.Create;
begin
inherited;
m_Arguments := tStringList .Create;
m_Options   := tStringLists.Create;				// owns the inserted stringlists
m_Switches  := tStringList .Create;  m_Switches.Sorted := True;
end;


destructor TCmdLine.Destroy;
begin
m_Arguments.Free;
m_Options  .Free;
m_Switches .Free;
inherited;
end;


function TCmdLine.Get_Argument(const Index : Word) : String;
begin
if (Index < m_Arguments.Count)	then Result := Arguments[Index]  // check if argument exists
				else Result := '';
end;


function TCmdLine.Get_Option(const Switch : String;  const Index : Word) : String;
var	i		:	Integer;
	Options		:	tStringList;
begin
i := m_Switches.IndexOf(Switch);
if (i <> -1) then begin						// check if switch exists
	Options := m_Options[i];
	if (Index < Options.Count) then begin			// check if option exists
		Result := Options[Index];
		Exit;
	end;
end;
Result := '';
end;


function TCmdLine.Get_OptionCount(const Switch : String) : Integer;
begin
Result := m_Switches.IndexOf(Switch);
if (Result = -1) then Exit;					// check if switch exists
Result := m_Options[Result].Count;
end;


function TCmdLine.has_Switch(const Switch : String) : Boolean;
begin
Result := (m_Switches.IndexOf(UpperCase(Switch)) <> -1);
end;


procedure TCmdLine.Parse;
var	i		:	Integer;
	InsertPos	:	Integer;
	p		:	Integer;
	Switch		:	String;
	tmpOptions	:	TStringList;
	tmpStr		:	String;
begin
for p := 1 to ParamCount do begin
	tmpStr := ParamStr(p);
	if (tmpStr[1] = '/') then begin				// [switch]
		Delete(tmpStr, 1, 1);				// remove switch indicator
		i := Pos('=', tmpStr);				// get switch & options
		if (i = 0) then i := MaxInt;
		Switch := UpperCase(Copy(tmpStr, 1, i - 1));
		InsertPos := m_Switches.Add(Switch);		// add switch at sorted position
		Delete(tmpStr, 1, i);				// get options
		tmpStr := StringReplace(tmpStr, ',', CRLF, [rfReplaceAll]);
		tmpOptions := tStringList.Create;		// insert options as string list
		tmpOptions.Text := tmpStr;
		m_Options.Insert(InsertPos, tmpOptions);
	end else begin
		m_Arguments.Add(tmpStr);				// [parameter]
	end;
end;
end;


///////////////////////////////////////////////////////////////////////////////////////////////


initialization
CmdLine := TCmdLine.Create;
CmdLine.Parse;							// parse right at program start


finalization
CmdLine.Free;


end.
