unit	UForm_Main;


interface
uses	Windows, Classes, ComCtrls, Controls, Dialogs, ExtCtrls, Forms, Graphics, ImgList,
	Messages, StdCtrls, SysUtils, Variants,
	UConsole, UINI, UShared, UTileDecoders;


type	TForm_Main	=	class(tForm)
				Btn_LoadPalette		:	TButton;
				Btn_LoadPicture		:	TButton;
				Btn_LoadTilemap		:	TButton;
				Btn_LoadVRAM		:	TButton;
				Btn_SavePalette		:	TButton;
				Btn_SaveTilemap		:	TButton;
				Btn_SaveVRAM		:	TButton;
				Dlg_LoadPalette		:	TOpenDialog;
				Dlg_LoadPicture		:	TOpenDialog;
				Dlg_LoadTilemap		:	TOpenDialog;
				Dlg_LoadVRAM		:	TOpenDialog;
				Dlg_SavePalette		:	TSaveDialog;
				Dlg_SaveTilemap		:	TSaveDialog;
				Dlg_SaveVRAM		:	TSaveDialog;
				GBox_Load		:	TGroupBox;
				GBox_Palette		:	TGroupBox;
				GBox_Save		:	TGroupBox;
				GBox_Tilemap		:	TGroupBox;
				GBox_Tiles		:	TGroupBox;
				Image_Palette		:	TImage;
				Image_Tilemap		:	TImage;
				ImageList_Tiles		:	TImageList;
				LV_Tiles		:	TListView;
				Panel_Load		:	TPanel;
				Panel_Palette		:	TPanel;
				Panel_Right		:	TPanel;
				Panel_Save		:	TPanel;
				Panel_Top		:	TPanel;
				ScrollBox		:	TScrollBox;
				StatusBar		:	TStatusBar;
				procedure	GUI_FormClose	(Sender : TObject;  var Action : TCloseAction);
				procedure	GUI_FormCreate	(Sender : TObject);
				procedure	GUI_FormKeyDown	(Sender : TObject;  var Key : Word;  Shift : TShiftState);
				procedure	GUI_LoadPalette	(Sender : TObject);
				procedure	GUI_LoadPicture	(Sender : TObject);
				procedure	GUI_LoadTilemap	(Sender : TObject);
				procedure	GUI_LoadVRAM	(Sender : TObject);
				procedure	GUI_MouseDown	(Sender : TObject;  Button : TMouseButton;  Shift : TShiftState;  x, y : Integer);
				procedure	GUI_MouseMove	(Sender : TObject;  Shift : TShiftState;  x, y : Integer);
				procedure	GUI_MouseUp	(Sender : TObject;  Button : TMouseButton;  Shift : TShiftState;  x, y : Integer);
				procedure	GUI_SavePalette	(Sender : TObject);
				procedure	GUI_SaveTilemap	(Sender : TObject);
				procedure	GUI_SaveVRAM	(Sender : TObject);
				private
				m_Palette		:	TPalette;
				m_Pos_MouseDown		:	TPoint;
				m_TileCount		:	Integer;
				m_TileMap		:	array[0..127, 0..127] of Byte;
				m_Tiles			:	array[Byte] of TTile;
				private
				procedure	Get_TilesFromBMP;
				procedure	Set_Palette	(const Bitmap : TBitmap);  overload;
				procedure	Set_Palette	(const Image  : TImage );  overload;
				procedure	Update_GUI	(const new_TileList : Boolean);
				procedure	Update_TileList	;
				procedure	Update_TileMap	;
				end;


var	Form_Main	:	TForm_Main;


implementation
uses	Math;


{$R *.dfm}


// tools


procedure Error(const Text : String);
begin
MessageDlg(Text, mtError, [mbOk], 0);
end;


procedure Init(const Bitmap : TBitmap;  const Size : Integer);
begin
with Bitmap do begin
	PixelFormat	:= pf8Bit;
	Width		:= Size;
	Height		:= Size;
	with Canvas do begin
		Brush.Color := clBlack;
		FillRect(ClipRect);
	end;
end;
Form_Main.Set_Palette(Bitmap);
end;


function is_SameTile(const Tile1, Tile2 : TTile) : Boolean;
var	x		:	Byte;
	y		:	Byte;
begin
for y := 0 to 7 do
for x := 0 to 7 do begin
	if (Tile1[y, x] <> Tile2[y, x]) then begin
		Result := False;
		Exit;
	end;
end;
Result := True;
end;


// TForm_Main


procedure TForm_Main.Get_TilesFromBMP;
var	i		:	Integer;
	j		:	Integer;
	Line		:	^TLine08;
	Tile		:	TTile;
	TileX		:	Integer;
	TileY		:	Integer;
	x, xPos		:	Integer;
	y, yPos		:	Integer;
begin
MemClear(m_Tiles, SizeOf(m_Tiles));
m_TileCount := 0;
for TileY := 0 to 127 do
for TileX := 0 to 127 do begin
	xPos := TileX * 8;					// get tile
	yPos := TileY * 8;
	with Image_Tilemap.Picture.Bitmap do begin
		for y := 0 to 7 do begin
			Line := ScanLine[yPos + y];
			for x := 0 to 7 do begin
				Tile[y, x] := Line[xPos + x];
			end;
		end;
	end;
	i := -1;						// get tile index
	for j := 0 to 255 do begin
		if is_SameTile(Tile, m_Tiles[j]) then begin
			i := j;
			break;
		end;
	end;
	if (i = -1) then begin					// unknown tile?
		if (m_TileCount < 256) then begin		// still room for new tiles?
			i := m_TileCount;
			m_Tiles[i] := Tile;
		end else begin
			i := 0;
		end;
		Inc(m_TileCount);
	end;
	m_TileMap[TileY, TileX] := i;				// update tilemap
end;
end;


procedure TForm_Main.GUI_FormClose(Sender: TObject; var Action: TCloseAction);
begin
with INI do begin
	Section := Self.Name;
	{Save_Form;}
	Save_FileDialog(Dlg_LoadPalette	);
	Save_FileDialog(Dlg_LoadPicture	);
	Save_FileDialog(Dlg_LoadTilemap	);
	Save_FileDialog(Dlg_LoadVRAM	);
	Save_FileDialog(Dlg_SavePalette	);
	Save_FileDialog(Dlg_SaveTilemap	);
	Save_FileDialog(Dlg_SaveVRAM	);
end;
end;


procedure TForm_Main.GUI_FormCreate(Sender : TObject);
var	BMP		:	TBitmap;
	i		:	Word;
	Line		:	pBytes;
	x		:	Byte;
	y		:	Byte;
begin
i := 0;								// startup colors
For y := 0 to $F do
For x := 0 to $F do Begin
	with m_Palette[i] do begin
		rgbRed   := ((y     ) SHL 1) SHL 3;
		rgbGreen := ((x     ) SHL 1) SHL 3;
		rgbBlue  := (($F - y) SHL 1) SHL 3;
	end;
	Inc(i);
end;
Image_Tilemap.Cursor := crHand;					// tilemap
Init(Image_Tilemap.Picture.Bitmap, 1024);
Init(Image_Palette.Picture.Bitmap,   16);			// palette
with Image_Palette.Picture.Bitmap do begin
	i := 0;
	for y := 0 to 15 do begin
		Line := ScanLine[y];
		for x := 0 to 15 do begin
			Line[x] := i;
			Inc(i);
		end;
	end;
end;
BMP := TBitmap.Create;						// tiles
try
	Init(BMP, 32);
	for i := 0 to 255 do  ImageList_Tiles.AddMasked(BMP, clNone);
finally
	BMP.Free;
end;
with LV_Tiles do begin
	for i := 0 to 255 do  with Items.Add do begin
		Caption := ' ' + IntToStr(i);
		ImageIndex := i;
	end;
	Columns[0].Width := -2;
	ItemFocused := Items[0];
	ItemFocused.Selected := True;
end;
Set_Buffering(Self);						// misc
ActiveControl := Btn_LoadPicture;
Caption := Application.Title;
with INI do begin
	Section := Self.Name;
	{Load_Form(False);}
	Load_FileDialog(Dlg_LoadPalette	);
	Load_FileDialog(Dlg_LoadPicture	);
	Load_FileDialog(Dlg_LoadTilemap	);
	Load_FileDialog(Dlg_LoadVRAM	);
	Load_FileDialog(Dlg_SavePalette	);
	Load_FileDialog(Dlg_SaveTilemap	);
	Load_FileDialog(Dlg_SaveVRAM	);
end;
end;


procedure TForm_Main.GUI_FormKeyDown(Sender : TObject;  var Key : Word;  Shift : TShiftState);
begin
Case Key of
	VK_Escape	:	Application.Terminate;
	else			Exit;
end;
Key := 0;
end;


procedure TForm_Main.GUI_LoadPalette(Sender : TObject);
var	i		:	Integer;
	Stream		:	TFileStream;
	Words		:	TWords256;
begin
if (not Exec_FileDlg(Dlg_LoadPalette)) then Exit;
Stream := TFileStream.Create(Dlg_FullPath, fmOpenRead OR fmShareDenyWrite);
try
	i := SizeOf(Words);
	if (Stream.Size < i) then begin
		Error('file too small');
		Exit;
	end;
	Stream.Read(Words, i);
finally
	Stream.Free;
end;
for i := 0 to 255 do  with m_Palette[i] do begin
	rgbRed   := ((Words[i] SHR 00) AND Bits5) SHL 3;
	rgbGreen := ((Words[i] SHR 05) AND Bits5) SHL 3;
	rgbBlue  := ((Words[i] SHR 10) AND Bits5) SHL 3;
end;
Update_GUI(True);
end;


procedure TForm_Main.GUI_LoadPicture(Sender : TObject);
var	BMP		:	TBitmap;
begin
if (not Exec_FileDlg(Dlg_LoadPicture)) then Exit;
BMP := TBitmap.Create;
try
	with BMP do begin
		LoadFromFile(Dlg_FullPath);
		if (not (PixelFormat in [pf1bit, pf4Bit, pf8Bit])) then begin
			Error('color depth must be 1, 4 or 8 bits per pixel');
			Exit;
		end;
		PixelFormat := pf8Bit;
		Canvas.Brush.Color := clBlack;
		Width  := 1024;
		Height := 1024;
		GetDIBColorTable(Canvas.Handle, 0, 256, m_Palette);
	end;
	Set_Palette(Image_Tilemap);
	Image_Tilemap.Canvas.Draw(0, 0, BMP);
finally
	BMP.Free;
end;
Get_TilesFromBMP;
Update_GUI(True);
end;


procedure TForm_Main.GUI_LoadTilemap(Sender : TObject);
var	Stream		:	TFileStream;
begin
if (not Exec_FileDlg(Dlg_LoadTilemap)) then Exit;
Stream := TFileStream.Create(Dlg_FullPath, fmOpenRead OR fmShareDenyWrite);
try
	if (Stream.Size < SizeOf(m_TileMap)) then begin
		Error('file too small');
		Exit;
	end;
	Stream.Read(m_TileMap, SizeOf(m_TileMap));
finally
	Stream.Free;
end;
Update_TileMap;
Update_GUI(False);
end;


procedure TForm_Main.GUI_LoadVRAM(Sender : TObject);
type	TVRAM_TileMap	=	array[0..127, 0..127, 0..1	] of Byte;
	TVRAM_Tiles	=	array[0..255			] of TPackedTile7i;
var	i		:	Byte;
	Stream		:	TFileStream;
	VRAM		:	TBytes64K;
	VRAM_TileMap	:	TVRAM_TileMap	absolute VRAM;
	VRAM_Tiles	:	TVRAM_Tiles	absolute VRAM;
	x		:	Byte;
	y		:	Byte;
begin
if (not Exec_FileDlg(Dlg_LoadVRAM)) then Exit;
Stream := TFileStream.Create(Dlg_FullPath, fmOpenRead OR fmShareDenyWrite);
try
	if (Stream.Size < SizeOf(VRAM)) then begin
		Error('file too small');
		Exit;
	end;
	Stream.Read(VRAM, SizeOf(VRAM));
finally
	Stream.Free;
end;
for y := 0 to 127 do						// tilemap (low bytes)
for x := 0 to 127 do  m_TileMap[y, x] := VRAM_TileMap[y, x, 0];
for i := 0 to 255 do begin					// tiles (high bytes)
	for y := 0 to 7 do
	for x := 0 to 7 do  m_Tiles[i][y, x] := VRAM_Tiles[i][y, x, 1];
end;
Update_TileMap;
Update_GUI(True);
end;


procedure TForm_Main.GUI_MouseDown(Sender : TObject;  Button : TMouseButton;  Shift : TShiftState;  x, y : Integer);
var	Image		:	TImage absolute Sender;
	tmpRect		:	TRect;
begin
Image .Cursor	:= crMove;
Screen.Cursor	:= crMove;
m_Pos_MouseDown	:= Point(x, y);
with tmpRect do begin
	Right	:= x + ScrollBox.HorzScrollBar.Position;
	Bottom	:= y + ScrollBox.VertScrollBar.Position;
	Left	:= Right  - (Image.Width  - ScrollBox.ClientWidth );
	Top	:= Bottom - (Image.Height - ScrollBox.ClientHeight);
	with m_Pos_MouseDown do begin
		If (Left > Right ) then begin  Left := x;  Right  := x;  end;
		If (Top  > Bottom) then begin  Top  := y;  Bottom := y;  end;
	end;
	Inc(Right );
	Inc(Bottom);
	TopLeft		:= Image.ClientToScreen(TopLeft    );
	BottomRight	:= Image.ClientToScreen(BottomRight);
end;
ClipCursor(@tmpRect);
end;


procedure TForm_Main.GUI_MouseMove(Sender : TObject;  Shift : TShiftState;  x, y : Integer);
var	Image		:	TImage absolute Sender;
	i		:	Integer;
begin
if (Image.Cursor = crMove) then begin
	with Image do  with ScrollBox do begin
		with HorzScrollBar do  Position := Position + (m_Pos_MouseDown.x - x);
		with VertScrollBar do  Position := Position + (m_Pos_MouseDown.y - y);
	end;
end else begin
	x := x div 8;
	y := y div 8;
	Assert(x >= 0);  Assert(x < 128);
	Assert(y >= 0);  Assert(y < 128);
	i := m_TileMap[y, x];
	with LV_Tiles do begin
		ItemFocused := Items[i];
		ItemFocused.Selected := True;
		ItemFocused.MakeVisible(False);
	end;
end;
end;


procedure TForm_Main.GUI_MouseUp(Sender : TObject;  Button : TMouseButton;  Shift : TShiftState;  x, y : Integer);
var	Image		:	TImage absolute Sender;
begin
with Image do  If (Cursor = crMove) then begin
	Cursor		:= crHand;
	Screen.Cursor	:= crDefault;
	ClipCursor(NIL);
end;
end;


procedure TForm_Main.GUI_SavePalette(Sender: TObject);
var	i		:	Integer;
	R, G, B		:	Integer;
	Stream		:	TFileStream;
	Words		:	TWords256;
begin
if (not Exec_FileDlg(Dlg_SavePalette)) then Exit;
for i := 0 to 255 do  with m_Palette[i] do begin
	R := rgbRed	SHR 3;
	G := rgbGreen	SHR 3;
	B := rgbBlue	SHR 3;
	Words[i] := (B SHL 10) OR (G SHL 5) OR (R)
end;
Stream := TFileStream.Create(Dlg_FullPath, fmOpenWrite OR fmShareExclusive);
try
	Stream.Write(Words, SizeOf(Words));
finally
	Stream.Free;
end;
end;


procedure TForm_Main.GUI_SaveTilemap(Sender: TObject);
var	Stream		:	TFileStream;
begin
if (not Exec_FileDlg(Dlg_SaveTilemap)) then Exit;
Stream := TFileStream.Create(Dlg_FullPath, fmOpenWrite OR fmShareExclusive);
try
	Stream.Write(m_TileMap, SizeOf(m_TileMap));
finally
	Stream.Free;
end;
end;


procedure TForm_Main.GUI_SaveVRAM(Sender : TObject);
type	TVRAM_TileMap	=	array[0..127, 0..127, 0..1	] of Byte;
	TVRAM_Tiles	=	array[0..255			] of TPackedTile7i;
var	i		:	Byte;
	Stream		:	TFileStream;
	VRAM		:	TBytes64K;
	VRAM_TileMap	:	TVRAM_TileMap	absolute VRAM;
	VRAM_Tiles	:	TVRAM_Tiles	absolute VRAM;
	x		:	Byte;
	y		:	Byte;
begin
if (not Exec_FileDlg(Dlg_SaveVRAM)) then Exit;
for y := 0 to 127 do						// tilemap (low bytes)
for x := 0 to 127 do  VRAM_TileMap[y, x, 0] := m_TileMap[y, x];
for i := 0 to 255 do begin					// tiles (high bytes)
	for y := 0 to 7 do
	for x := 0 to 7 do  VRAM_Tiles[i][y, x, 1] := m_Tiles[i][y, x];
end;
Stream := TFileStream.Create(Dlg_FullPath, fmOpenWrite OR fmShareExclusive);
try
	Stream.Write(VRAM, SizeOf(VRAM));
finally
	Stream.Free;
end;
end;


procedure TForm_Main.Set_Palette(const Bitmap : TBitmap);
begin
SetDIBColorTable(Bitmap.Canvas.Handle, 0, 256, m_Palette);
end;


procedure TForm_Main.Set_Palette(const Image : TImage);
begin
with Image do begin
	Set_Palette(Picture.Bitmap);
	Invalidate;
end;
end;


procedure TForm_Main.Update_GUI(const new_TileList : Boolean);
begin
Set_Palette(Image_Tilemap);
Set_Palette(Image_Palette);
if new_TileList then Update_TileList;
StatusBar.SimpleText := ' tiles: ' + IntToStr(m_TileCount);
end;


procedure TForm_Main.Update_TileList;
var	BMP1		:	TBitmap;
	BMP2		:	TBitmap;
	i		:	Byte;
	Line		:	^TLine08;
	x		:	Byte;
	y		:	Byte;
begin
ImageList_Tiles.Clear;
BMP1 := TBitmap.Create;
BMP2 := TBitmap.Create;
try
	Init(BMP1, 08);  Set_Palette(BMP1);
	Init(BMP2, 32);  Set_Palette(BMP2);
	for i := 0 to 255 do begin
		for y := 0 to 7 do begin
			Line := BMP1.ScanLine[y];
			for x := 0 to 7 do begin
				Line[x] := m_Tiles[i][y, x];
			end;
		end;
		with BMP2.Canvas do begin
			StretchDraw(ClipRect, BMP1);
		end;
		ImageList_Tiles.AddMasked(BMP2, clNone);
	end;
finally
	BMP1.Free;
	BMP2.Free;
end;
LV_Tiles.Invalidate;
end;


procedure TForm_Main.Update_TileMap;
var	i		:	Integer;
	Line		:	^TLine08;
	Tile		:	^TTile;
	TileX		:	Integer;
	TileY		:	Integer;
	x, xPos		:	Integer;
	y, yPos		:	Integer;
begin
with Image_Tilemap do begin
	with Picture.Bitmap do begin
		for TileY := 0 to 127 do
		for TileX := 0 to 127 do begin
			i := m_TileMap[TileY, TileX];
			Tile := @m_Tiles[i];
			xPos := TileX * 8;
			yPos := TileY * 8;
			for y := 0 to 7 do begin
				Line := ScanLine[yPos + y];
				for x := 0 to 7 do begin
					Line[xPos + x] := Tile[y, x];
				end;
			end;
		end;
	end;
	Invalidate;
end;
end;


end.
